from collections.abc import Iterator, Sequence
from typing import Protocol

from mypy.types import (
    AnyType,
    Instance,
    ProperType,
    TypeOfAny,
    TypeType,
    TypeVarType,
    UnboundType,
    UnionType,
    get_proper_type,
)
from mypy.types import Type as MypyType
from typing_extensions import assert_never

from .. import _known_annotations, _store


class FailFunc(Protocol):
    def __call__(self, msg: str) -> None: ...


class DeferFunc(Protocol):
    def __call__(self) -> bool: ...


class TypeAnalyze(Protocol):
    def __call__(self, typ: MypyType, /) -> MypyType: ...


class NamedTypeOrNone(Protocol):
    def __call__(self, fullname: str, args: list[MypyType] | None = None) -> Instance | None: ...


class AnnotationResolver:
    def __init__(
        self,
        store: _store.Store,
        fail: FailFunc,
        defer: DeferFunc,
        lookup_info: _store.LookupInfo,
        named_type_or_none: NamedTypeOrNone,
    ) -> None:
        self._store = store
        self._fail = fail
        self._defer = defer
        self._named_type_or_none = named_type_or_none
        self._lookup_info = lookup_info

    def _flatten_union(self, typ: ProperType) -> Iterator[ProperType]:
        if isinstance(typ, UnionType):
            for item in typ.items:
                yield from self._flatten_union(get_proper_type(item))
        else:
            yield typ

    def _analyze_first_type_arg(
        self, type_arg: ProperType
    ) -> tuple[bool, Sequence[Instance] | None]:
        is_type: bool = False

        found: ProperType = type_arg
        if isinstance(type_arg, TypeType):
            is_type = True
            found = type_arg.item

        if isinstance(found, AnyType):
            self._fail("Tried to use concrete annotations on a typing.Any")
            return False, None

        if not isinstance(found, Instance | UnionType):
            return False, None

        if isinstance(found, Instance):
            found = UnionType((found,))

        all_types = list(self._flatten_union(found))
        all_instances: list[Instance] = []
        not_all_instances: bool = False
        for item in all_types:
            if not isinstance(item, Instance):
                self._fail(
                    f"Expected to operate on specific classes, got a {item.__class__.__name__}: {item}"
                )
                not_all_instances = True
            else:
                all_instances.append(item)

        if not_all_instances:
            return False, None

        concrete: list[Instance] = []
        names = ", ".join([item.type.fullname for item in all_instances])

        for item in all_instances:
            concrete.extend(
                self._store.retrieve_concrete_children_types(
                    item.type, self._lookup_info, self._named_type_or_none
                )
            )

        if not concrete:
            if not self._defer():
                self._fail(f"No concrete models found for {names}")
            return False, None

        return is_type, tuple(concrete)

    def _make_union(
        self, is_type: bool, instances: Sequence[Instance]
    ) -> UnionType | Instance | TypeType:
        items: Sequence[UnionType | TypeType | Instance]

        if is_type:
            items = [item if isinstance(item, TypeType) else TypeType(item) for item in instances]
        else:
            items = instances

        if len(items) == 1:
            return items[0]
        else:
            return UnionType(tuple(items))

    def resolve(
        self, annotation: _known_annotations.KnownAnnotations, type_arg: ProperType
    ) -> Instance | TypeType | UnionType | AnyType | None:
        if annotation is _known_annotations.KnownAnnotations.CONCRETE:
            return self.find_concrete_models(type_arg)
        elif annotation is _known_annotations.KnownAnnotations.DEFAULT_QUERYSET:
            return self.find_default_queryset(type_arg)
        else:
            assert_never(annotation)

    def find_type_arg(
        self, unbound_type: UnboundType, analyze_type: TypeAnalyze
    ) -> tuple[ProperType | None, bool]:
        args = unbound_type.args
        if len(args := unbound_type.args) != 1:
            self._fail("Concrete annotations must contain exactly one argument")
            return None, False

        type_arg = get_proper_type(analyze_type(args[0]))
        needs_rewrap = self._has_typevars(type_arg)
        return type_arg, needs_rewrap

    def _has_typevars(self, type_arg: ProperType) -> bool:
        if isinstance(type_arg, TypeType):
            type_arg = type_arg.item

        if isinstance(type_arg, TypeVarType):
            return True

        if not isinstance(type_arg, UnionType):
            return False

        return any(self._has_typevars(get_proper_type(item)) for item in type_arg.items)

    def find_concrete_models(
        self, type_arg: ProperType
    ) -> Instance | TypeType | UnionType | AnyType | None:
        is_type, concrete = self._analyze_first_type_arg(type_arg)
        if concrete is None:
            return None

        return self._make_union(is_type, concrete)

    def find_default_queryset(
        self, type_arg: ProperType
    ) -> Instance | TypeType | UnionType | AnyType | None:
        is_type, concrete = self._analyze_first_type_arg(type_arg)
        if concrete is None:
            return None

        try:
            querysets = tuple(
                self._store.realise_querysets(UnionType(concrete), self._lookup_info)
            )
        except _store.RestartDmypy as err:
            self._fail(f"You probably need to restart dmypy: {err}")
            return AnyType(TypeOfAny.from_error)
        except _store.UnionMustBeOfTypes:
            self._fail("Union must be of instances of models")
            return None
        else:
            return self._make_union(is_type, querysets)



import textwrap
from collections.abc import Mapping
from typing import Protocol, TypedDict, overload

from pytest_mypy_plugins import (
    File,
    FollowupFile,
    MypyPluginsConfig,
    MypyPluginsScenario,
    OutputChecker,
)
from pytest_mypy_plugins.scenario import Strategy
from typing_extensions import NotRequired, Unpack

from .output_builder import OutputBuilder


class RunArgs(TypedDict):
    start: NotRequired[list[str]]
    expect_fail: NotRequired[bool]
    additional_properties: NotRequired[Mapping[str, object]]
    copied_apps: NotRequired[list[str]]
    installed_apps: NotRequired[list[str]]
    debug: NotRequired[bool]
    OutputCheckerKls: NotRequired[type[OutputChecker]]


class Action(Protocol):
    def __call__(self, expected: OutputBuilder) -> None: ...


class _Runner:
    def __init__(
        self, scenario: "Scenario", output_builder: OutputBuilder, **kwargs: Unpack[RunArgs]
    ) -> None:
        self.scenario = scenario
        self.output_builder = output_builder
        self.run_kwargs = kwargs

    def __call__(self, action: Action, /) -> None:
        action(self.output_builder)
        return self.scenario.run_and_check_mypy(self.output_builder, **self.run_kwargs)


class Scenario:
    def __init__(self, config: MypyPluginsConfig, scenario: MypyPluginsScenario) -> None:
        self.config = config
        self.scenario = scenario
        self.ran_at_least_once: bool = False
        self.for_daemon = self.config.strategy is Strategy.DAEMON
        self.expected = OutputBuilder(for_daemon=self.for_daemon)

    def file(self, expected: OutputBuilder, path: str, content: str | None) -> None:
        content = expected.parse_content(path, content)

        if self.ran_at_least_once or content is None:
            assert self.ran_at_least_once
            followup = FollowupFile(path=path, content=content)
            self.scenario.handle_followup_file(followup)
        else:
            file = File(path=path, content=content)
            self.scenario.make_file(file)

    def append_to_file(self, expected: OutputBuilder, path: str, content: str) -> None:
        assert self.ran_at_least_once

        location = self.scenario.execution_path / path
        assert location.exists()
        file = FollowupFile(path=path, content=location.read_text() + textwrap.dedent(content))
        self.scenario.handle_followup_file(file)

    def run_and_check_mypy(
        self, expected_output: OutputBuilder, **kwargs: Unpack[RunArgs]
    ) -> None:
        start = kwargs.get("start", ["."])

        extra_properties: dict[str, object] = {}

        if "installed_apps" in kwargs:
            extra_properties["installed_apps"] = kwargs["installed_apps"]

        if "debug" in kwargs:
            extra_properties["debug"] = kwargs["debug"]

        if "copied_apps" in kwargs:
            extra_properties["copied_apps"] = kwargs["copied_apps"]

        try:
            return self.scenario.run_and_check_mypy(
                start,
                expect_fail=kwargs.get("expect_fail", False),
                expected_output=list(expected_output),
                additional_properties={
                    **kwargs.get("additional_properties", {}),
                    **extra_properties,
                },
                OutputCheckerKls=kwargs.get("OutputCheckerKls", OutputChecker),
            )
        finally:
            self.ran_at_least_once = True

    @overload
    def run_and_check_mypy_after(
        self, action: None = None, /, **kwargs: Unpack[RunArgs]
    ) -> _Runner: ...

    @overload
    def run_and_check_mypy_after(self, action: Action, /, **kwargs: Unpack[RunArgs]) -> None: ...

    def run_and_check_mypy_after(
        self, action: Action | None = None, /, **kwargs: Unpack[RunArgs]
    ) -> _Runner | None:
        runner = _Runner(self, self.expected, **kwargs)
        if action is None:
            return runner
        else:
            runner(action)
            return None



from django.db import models
from leader.models import Leader


class Follower1QuerySet(models.QuerySet["Follower1"]):
    def good_ones(self) -> "Follower1QuerySet":
        return self.filter(good=True)


Follower1Manager = models.Manager.from_queryset(Follower1QuerySet)


class Follower1(Leader):
    good = models.BooleanField()
    from_follower1 = models.CharField()

    objects = Follower1Manager()



import os
import pathlib
import textwrap

import pytest
import pytest_mypy_plugins.utils
from extended_mypy_django_plugin_test_driver import OutputBuilder, Scenario, assertions
from pytest_mypy_plugins import OutputChecker


class TestErrors:
    def test_cant_use_typevar_concrete_annotation_in_function_or_method_typeguard(
        self, scenario: Scenario
    ) -> None:
        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            scenario.file(
                expected,
                "main.py",
                """
                from typing import TypeGuard, TypeVar, cast, TypeVar

                from myapp.models import Child1, Parent

                from extended_mypy_django_plugin import Concrete

                T_Parent = TypeVar("T_Parent", bound=Parent)

                def function_with_type_typeguard(
                    cls: type[T_Parent]
                ) -> TypeGuard[type[Concrete[T_Parent]]]:
                    return hasattr(cls, "objects")

                cls1: type[Parent] = Child1
                assert function_with_type_typeguard(cls1)
                # ^ ERROR(misc) ^ Can't use a TypeGuard that uses a Concrete Annotation that uses type variables
                cls1
                # ^ REVEAL ^ type[extended_mypy_django_plugin.annotations.Concrete[myapp.models.Parent]]

                def function_with_instance_typeguard(
                    instance: T_Parent
                ) -> TypeGuard[Concrete[T_Parent]]:
                    return True

                instance1: Parent = cast(Child1, None)
                assert function_with_instance_typeguard(instance1)
                # ^ ERROR(misc) ^ Can't use a TypeGuard that uses a Concrete Annotation that uses type variables
                instance1
                # ^ REVEAL ^ extended_mypy_django_plugin.annotations.Concrete[myapp.models.Parent]

                class Logic:
                    def method_with_type_typeguard(
                        self, cls: type[T_Parent]
                    ) -> TypeGuard[type[Concrete[T_Parent]]]:
                        return hasattr(cls, "objects")

                    def method_with_instance_typeguard(
                        self, instance: T_Parent
                    ) -> TypeGuard[Concrete[T_Parent]]:
                        return True

                logic = Logic()
                cls2: type[Parent] = Child1
                assert logic.method_with_type_typeguard(cls2)
                # ^ ERROR(misc) ^ Can't use a TypeGuard that uses a Concrete Annotation that uses type variables
                cls2
                # ^ REVEAL ^ type[extended_mypy_django_plugin.annotations.Concrete[T_Parent`-1]]

                instance2: Parent = cast(Child1, None)
                assert logic.method_with_instance_typeguard(instance2)
                # ^ ERROR(misc) ^ Can't use a TypeGuard that uses a Concrete Annotation that uses type variables
                instance2
                # ^ REVEAL ^ extended_mypy_django_plugin.annotations.Concrete[T_Parent`-1]
                """,
            )

    def test_gracefully_handles_determine_version_failure_on_startup(
        self, scenario: Scenario, tmp_path: pathlib.Path
    ) -> None:
        if not scenario.for_daemon:
            pytest.skip("Test only relevant for the daemon")

        determine_script = tmp_path / "determine.py"

        determine_script.write_text(
            textwrap.dedent("""
        #!/usr/bin/env python

        from extended_mypy_django_plugin.scripts.determine_django_state import main


        if __name__ == "__main__":
            raise ValueError("Computer says no")
            main()
        """)
        )

        os.chmod(determine_script, 0o755)
        scenario.scenario.additional_mypy_config += (
            f"\ndetermine_django_state_script = {determine_script}"
        )

        with pytest.raises(pytest_mypy_plugins.utils.TypecheckAssertionError) as err:

            @scenario.run_and_check_mypy_after
            def _(expected: OutputBuilder) -> None:
                pass

        assert err.value.mypy_output is not None

        assertions.assert_glob_lines(
            err.value.mypy_output,
            f"""
            Error constructing plugin instance of Plugin
            
            Daemon crashed!
            Traceback (most recent call last):
            RuntimeError:
            Failed to determine information about the django setup
            
              > *python {determine_script} --django-settings-module mysettings --apps-file * --known-models-file * --scratch-path *
              |
              | Traceback (most recent call last):
              |   File "{determine_script}", line 8, in <module>
              |     raise ValueError("Computer says no")
              | ValueError: Computer says no
              |
            """,
        )

    def test_gracefully_handles_determine_version_failure_on_subsequent_run(
        self, scenario: Scenario, tmp_path: pathlib.Path
    ) -> None:
        if not scenario.for_daemon:
            pytest.skip("Test only relevant for the daemon")

        determine_script = tmp_path / "determine.py"

        determine_script.write_text(
            textwrap.dedent("""
        #!/usr/bin/env python

        from extended_mypy_django_plugin.scripts.determine_django_state import main


        if __name__ == "__main__":
            main()
        """)
        )

        os.chmod(determine_script, 0o755)
        scenario.scenario.additional_mypy_config += (
            f"\ndetermine_django_state_script = {determine_script}"
        )

        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            pass

        determine_script.write_text(
            textwrap.dedent("""
        #!/usr/bin/env python

        from extended_mypy_django_plugin.scripts.determine_django_state import main


        if __name__ == "__main__":
            raise ValueError("Computer says no")
            main()
        """)
        )

        called: list[int] = []

        class CheckNoCrashShowsFailure(OutputChecker):
            def check(self, ret_code: int, stdout: str, stderr: str) -> None:
                called.append(ret_code)

                assert ret_code == 0
                assertions.assert_glob_lines(
                    stdout + stderr,
                    f"""
                    Failed to determine information about the django setup
                    
                    > */python {determine_script} --django-settings-module mysettings --apps-file * --known-models-file * --scratch-path *
                    |
                    | Traceback (most recent call last):
                    |   File "{determine_script}", line 8, in <module>
                    |     raise ValueError("Computer says no")
                    | ValueError: Computer says no
                    |
                    """,
                )

        scenario.run_and_check_mypy(scenario.expected, OutputCheckerKls=CheckNoCrashShowsFailure)
        assert called == [0]

        determine_script.write_text(
            textwrap.dedent("""
        #!/usr/bin/env python

        from extended_mypy_django_plugin.scripts.determine_django_state import main


        if __name__ == "__main__":
            main()
        """)
        )

        class CheckNoOutput(OutputChecker):
            def check(self, ret_code: int, stdout: str, stderr: str) -> None:
                called.append(ret_code)

                assert ret_code == 0
                assert stdout + stderr == ""

        scenario.run_and_check_mypy(scenario.expected, OutputCheckerKls=CheckNoOutput)
        assert called == [0, 0]



import fnmatch


def assert_glob_lines(got: str, expect: str) -> None:
    message = [line.strip() for line in got.strip().replace("\r\n", "\n").split("\n")]
    want = [line.strip() for line in expect.strip().split("\n")]

    print("GOT >>" + "=" * 74)
    print()
    print("\n".join(message))
    print()
    print("WANT >>" + "-" * 73)
    print()
    print("\n".join(want))

    count = 1
    while want:
        line = want[0]
        if not message:
            assert False, f"Ran out of lines, stopped at [{count}] '{want[0]}'"

        if message[0] == line or fnmatch.fnmatch(message[0], line):
            count += 1
            want.pop(0)

        message.pop(0)

    if want:
        assert False, f"Didn't match all the lines, stopped at [{count}] '{want[0]}'"



from extended_mypy_django_plugin_test_driver import OutputBuilder, Scenario


def test_works(scenario: Scenario) -> None:
    @scenario.run_and_check_mypy_after
    def _(expected: OutputBuilder) -> None:
        scenario.file(
            expected,
            "main.py",
            """
            from extended_mypy_django_plugin import Concrete, DefaultQuerySet

            from myapp.models import Parent, Child1, Child2

            T_Child = Concrete.type_var("T_Child", Parent)


            def make_child(child: type[T_Child]) -> T_Child:
                return child.objects.create()


            def make_any_queryset(child: type[Concrete[Parent]]) -> DefaultQuerySet[Parent]:
                return child.objects.all()


            def make_child1_queryset() -> DefaultQuerySet[Child1]:
                return Child1.objects.all()


            def make_child2_queryset() -> DefaultQuerySet[Child2]:
                return Child2.objects.all()


            def make_multiple_queryset(child: type[Child1 | Child2]) -> DefaultQuerySet[Child2 | Child1]:
                return child.objects.all()


            def make_child_typevar_queryset(child: type[T_Child]) -> DefaultQuerySet[T_Child]:
                return child.objects.all()


            def ones(model: type[Concrete[Parent]]) -> list[str]:
                model.objects
                # ^ REVEAL ^ Union[django.db.models.manager.Manager[myapp.models.Child1], myapp.models.ManagerFromChild2QuerySet[myapp.models.Child2], django.db.models.manager.Manager[myapp.models.Child3], django.db.models.manager.Manager[myapp2.models.ChildOther]]
                return list(model.objects.values_list("one", flat=True))


            make_child(Child1)
            # ^ REVEAL ^ myapp.models.Child1

            make_any_queryset(Child1)
            # ^ REVEAL ^ Union[django.db.models.query.QuerySet[myapp.models.Child1, myapp.models.Child1], myapp.models.Child2QuerySet, django.db.models.query.QuerySet[myapp.models.Child3, myapp.models.Child3], django.db.models.query.QuerySet[myapp2.models.ChildOther, myapp2.models.ChildOther]]

            make_child1_queryset()
            # ^ REVEAL ^ django.db.models.query.QuerySet[myapp.models.Child1, myapp.models.Child1]

            qs2 = make_child2_queryset()
            # ^ REVEAL ^ myapp.models.Child2QuerySet

            qs2.all()
            # ^ REVEAL ^ myapp.models.Child2QuerySet

            Child2.objects
            # ^ REVEAL ^ myapp.models.ManagerFromChild2QuerySet[myapp.models.Child2]

            Child2.objects.all()
            # ^ REVEAL ^ myapp.models.Child2QuerySet[myapp.models.Child2]

            make_child_typevar_queryset(Child1)
            # ^ REVEAL ^ django.db.models.query.QuerySet[myapp.models.Child1, myapp.models.Child1]

            make_child_typevar_queryset(Child2)
            # ^ REVEAL ^ myapp.models.Child2QuerySet

            make_multiple_queryset(Child1)
            # ^ REVEAL ^ Union[myapp.models.Child2QuerySet, django.db.models.query.QuerySet[myapp.models.Child1, myapp.models.Child1]]
            """,
        )



from .find_models import find_known_models, record_known_models
from .find_settings import find_known_settings, record_known_settings

__all__ = [
    "record_known_models",
    "find_known_models",
    "find_known_settings",
    "record_known_settings",
]



import pathlib
from collections.abc import Iterator
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from django.conf import LazySettings


def record_known_settings(output_location: pathlib.Path, settings: "LazySettings") -> None:
    found: set[str] = set()
    for line in find_known_settings(settings):
        found.add(line)

    output_location.write_text("\n".join(sorted(found)))


def find_known_settings(settings: "LazySettings") -> Iterator[str]:
    for name in dir(settings):
        if name.startswith("_"):
            continue

        yield f"{name} :: {type(getattr(settings, name))}"



import collections
import dataclasses
import enum
import importlib.metadata
import re
import textwrap
from collections.abc import Iterator, Sequence
from itertools import chain

from pytest_mypy_plugins import OutputMatcher
from pytest_mypy_plugins.utils import (
    DaemonOutputMatcher,
    FileOutputMatcher,
    extract_output_matchers_from_out,
)
from typing_extensions import Self, assert_never

regexes = {
    "potential_instruction": re.compile(r"^\s*#\s*\^"),
    "instruction": re.compile(
        r"^(?P<prefix_whitespace>\s*)#\s*\^\s*(?P<instruction>TAG|REVEAL|ERROR|NOTE)(\((?P<options>[^\)]*)\))?(\[(?P<tag>[^\]]*)\])?\s*\^\s*(?P<rest>.*)"
    ),
    "assignment": re.compile(r"^(?P<var_name>[a-zA-Z0-9_]+)\s*(:[^=]+)?(=|$)"),
}


class _Instruction(enum.Enum):
    TAG = "TAG"
    REVEAL = "REVEAL"
    ERROR = "ERROR"
    NOTE = "NOTE"


class _Build:
    @dataclasses.dataclass
    class _File:
        tags: dict[str, FileOutputMatcher] = dataclasses.field(default_factory=dict)
        matchers: list[FileOutputMatcher] = dataclasses.field(default_factory=list)

        def clear(self) -> None:
            self.tags.clear()
            self.matchers.clear()

    def __init__(self, for_daemon: bool) -> None:
        self.for_daemon = for_daemon
        self._by_file: dict[str, _Build._File] = collections.defaultdict(lambda: _Build._File())
        self.daemon_should_restart: bool = False

    @property
    def result(self) -> Sequence[OutputMatcher]:
        return list(chain.from_iterable([f.matchers for f in self._by_file.values()]))

    def matcher_for_tag(self, path: str, tag: str) -> FileOutputMatcher:
        if path not in self._by_file or tag not in self._by_file[path].tags:
            raise AssertionError(f"No previously made matcher for {path} with tag {tag}")

        return self._by_file[path].tags[tag]

    def remove_tagged(self, path: str, tag: str, leave_tag: bool = False) -> None:
        lnum = self.matcher_for_tag(path, tag).lnum
        self._by_file[path].matchers = [
            matcher for matcher in self._by_file[path].matchers if matcher.lnum != lnum
        ]
        if not leave_tag:
            del self._by_file[path].tags[tag]

    def add(
        self,
        path: str,
        lnum: int,
        col: int | None,
        severity: str,
        message: str,
        regex: bool = False,
        tag: str = "",
        tag_only: bool = False,
        override_tag: bool = False,
    ) -> None:
        fname = path.removesuffix(".py")
        matcher = FileOutputMatcher(
            fname, lnum, severity, message, regex=regex, col=None if col is None else str(col)
        )

        for_file = self._by_file[path]

        existing_matcher: FileOutputMatcher | None = None
        if tag and override_tag and tag in for_file.tags:
            existing_matcher = for_file.tags[tag]

        if not tag_only:
            if tag and override_tag and tag in for_file.tags:
                for_file.matchers = [
                    matcher for matcher in for_file.matchers if matcher is not existing_matcher
                ]

            for_file.matchers.append(matcher)

        if tag:
            if not override_tag:
                assert (
                    tag not in for_file.tags
                ), f"Already have a matched tagged as {tag} for {path}"
            for_file.tags[tag] = matcher

    def clear_path(self, path: str) -> None:
        self._by_file[path].clear()

    def clear(self) -> None:
        self._by_file.clear()


class OutputBuilder:
    def __init__(
        self,
        build: _Build | None = None,
        target_file: str | None = None,
        for_daemon: bool | None = False,
    ) -> None:
        if build is None:
            assert for_daemon is not None
            build = _Build(for_daemon=for_daemon)
        self._build = build

        self.target_file = target_file

    def _normalise_message(self, message: str) -> str:
        if importlib.metadata.version("mypy") == "1.4.0":
            return message.replace("type[", "Type[").replace(
                "django.db.models.query.QuerySet", "django.db.models.query._QuerySet"
            )
        else:
            return message

    def _split_lnum_or_tag(self, lnum_or_tag: int | str) -> tuple[int, str]:
        if isinstance(lnum_or_tag, int):
            return lnum_or_tag, ""
        else:
            return -1, lnum_or_tag

    def clear(self) -> Self:
        if self.target_file:
            self._build.clear_path(self.target_file)
        else:
            self._build.clear()
        return self

    def daemon_should_restart(self) -> Self:
        self._build.daemon_should_restart = True
        return self

    def daemon_should_not_restart(self) -> Self:
        self._build.daemon_should_restart = False
        return self

    def on(self, path: str) -> Self:
        return self.__class__(build=self._build, target_file=path)

    def from_out(self, out: str, regex: bool = False) -> Self:
        if importlib.metadata.version("mypy") == "1.4.0":
            out = out.replace("type[", "Type[").replace(
                "django.db.models.query.QuerySet", "django.db.models.query._QuerySet"
            )

        for matcher in extract_output_matchers_from_out(
            out, {}, regex=regex, for_daemon=self._build.for_daemon
        ):
            assert isinstance(matcher, FileOutputMatcher)
            self._build._by_file[f"{matcher.fname}.py"].matchers.append(matcher)
        return self

    def add_revealed_type(self, lnum_or_tag: int | str, revealed_type: str, tag: str = "") -> Self:
        assert self.target_file is not None

        revealed_type = self._normalise_message(revealed_type)

        lnum, itag = self._split_lnum_or_tag(lnum_or_tag)

        if itag:
            if not tag:
                tag = itag
            lnum = self._build.matcher_for_tag(self.target_file, tag).lnum

        self._build.add(
            self.target_file,
            lnum,
            None,
            "note",
            f'Revealed type is "{revealed_type}"',
            tag=tag,
            override_tag=True,
        )
        return self

    def remove_errors(self, lnum_or_tag: int | str, leave_tag: bool = False) -> Self:
        assert self.target_file is not None

        lnum, tag = self._split_lnum_or_tag(lnum_or_tag)

        if tag:
            self._build.remove_tagged(self.target_file, tag, leave_tag=leave_tag)
            return self

        self._build._by_file[self.target_file].matchers = [
            matcher
            for matcher in self._build._by_file[self.target_file].matchers
            if matcher.lnum != lnum and matcher.severity != "error"
        ]

        return self

    def add_error(
        self, lnum_or_tag: int | str, error_type: str, message: str, tag: str = ""
    ) -> Self:
        assert self.target_file is not None

        message = self._normalise_message(message)

        lnum, itag = self._split_lnum_or_tag(lnum_or_tag)

        if itag:
            if not tag:
                tag = itag
            lnum = self._build.matcher_for_tag(self.target_file, tag).lnum

        self._build.add(
            self.target_file,
            lnum,
            None,
            "error",
            f"{message}  [{error_type}]",
            tag=tag,
            override_tag=True,
        )
        return self

    def replace_errors(self, tag: str, *errors: tuple[str, str]) -> Self:
        assert self.target_file is not None

        lnum = self._build.matcher_for_tag(self.target_file, tag).lnum
        self.remove_errors(tag, leave_tag=True)

        for error_type, message in errors:
            self.add_error(lnum, error_type, message)

        return self

    def add_note(self, lnum_or_tag: int | str, message: str, tag: str = "") -> Self:
        assert self.target_file is not None

        message = self._normalise_message(message)

        lnum, itag = self._split_lnum_or_tag(lnum_or_tag)

        if itag:
            if not tag:
                tag = itag
            lnum = self._build.matcher_for_tag(self.target_file, tag).lnum

        self._build.add(self.target_file, lnum, None, "note", message, tag=tag, override_tag=True)
        return self

    def remove_from_revealed_type(self, lnum_or_tag: int | str, remove: str) -> Self:
        assert self.target_file is not None

        remove = self._normalise_message(remove)

        lnum, tag = self._split_lnum_or_tag(lnum_or_tag)

        if tag:
            lnum = self._build.matcher_for_tag(self.target_file, tag).lnum

        found: list[FileOutputMatcher] = []
        for matcher in self._build._by_file[self.target_file].matchers:
            if (
                matcher.lnum == lnum
                and matcher.severity == "note"
                and matcher.message.startswith("Revealed type is")
            ):
                found.append(matcher)

        assert len(found) == 1
        found[0].message = found[0].message.replace(remove, "")
        return self

    def parse_content(self, path: str, content: str | None) -> str | None:
        self._build.clear_path(path)

        if content is None:
            return content

        content = textwrap.dedent(content).lstrip()
        result: list[str] = []
        expected = self.on(path)

        lines = content.split("\n")
        for i, line in enumerate(lines):
            m = regexes["instruction"].match(line)
            if m is None:
                if regexes["potential_instruction"].match(line):
                    raise AssertionError(
                        f"Looks like line is trying to be an expectation but it didn't pass the regex for one: {line}"
                    )
                result.append(line)
                continue

            gd = m.groupdict()
            result.append("")
            expected._parse_instruction(
                i,
                result,
                line,
                prefix_whitespace=gd["prefix_whitespace"],
                instruction=_Instruction(gd["instruction"]),
                options=gd.get("options", "") or "",
                tag=gd.get("tag", "") or "",
                rest=gd["rest"],
            )

        return "\n".join(result)

    def _parse_instruction(
        self,
        i: int,
        result: list[str],
        line: str,
        *,
        prefix_whitespace: str,
        instruction: _Instruction,
        options: str,
        tag: str,
        rest: str,
    ) -> None:
        if instruction is _Instruction.REVEAL:
            previous_line = result[i - 1]
            m = regexes["assignment"].match(previous_line.strip())
            if m:
                result[i] = f"{prefix_whitespace}reveal_type({m.groupdict()['var_name']})"
                i += 1
            else:
                result[i - 1] = f"{prefix_whitespace}reveal_type({previous_line.strip()})"

            self.add_revealed_type(i, rest, tag=tag)
        elif instruction is _Instruction.ERROR:
            assert options, "Must use `# ^ ERROR(error-type) ^ error here`"
            self.add_error(i, options, rest, tag=tag)
        elif instruction is _Instruction.NOTE:
            self.add_note(i, rest, tag=tag)
        elif instruction is _Instruction.TAG:
            assert tag, "Must use a `# ^ TAG[tag-name] ^`"
            assert self.target_file is not None
            self._build.add(self.target_file, i, None, "error", "", tag=tag, tag_only=True)
        else:
            assert_never(instruction)

    def __iter__(self) -> Iterator[OutputMatcher]:
        if self._build.daemon_should_restart and self._build.for_daemon:
            yield DaemonOutputMatcher(line="Restarting: plugins changed", regex=False)
            yield DaemonOutputMatcher(line="Daemon stopped", regex=False)
        yield from self._build.result



from ._annotation_resolver import AnnotationResolver
from ._sem_analyze import SemAnalyzing, TypeAnalyzer
from ._type_checker import SharedCheckTypeGuardsLogic, SharedModifyReturnTypeLogic, TypeChecking

__all__ = [
    "SemAnalyzing",
    "TypeAnalyzer",
    "AnnotationResolver",
    "TypeChecking",
    "SharedCheckTypeGuardsLogic",
    "SharedModifyReturnTypeLogic",
]



import enum


class KnownClasses(enum.Enum):
    CONCRETE = "extended_mypy_django_plugin.annotations.Concrete"


class KnownAnnotations(enum.Enum):
    CONCRETE = "extended_mypy_django_plugin.annotations.Concrete"
    DEFAULT_QUERYSET = "extended_mypy_django_plugin.annotations.DefaultQuerySet"



import enum
import functools
import sys
from collections.abc import Callable
from typing import Generic

from mypy.checker import TypeChecker
from mypy.modulefinder import mypy_path
from mypy.nodes import MypyFile, SymbolNode, SymbolTableNode, TypeInfo
from mypy.options import Options
from mypy.plugin import (
    AnalyzeTypeContext,
    AttributeContext,
    DynamicClassDefContext,
    FunctionContext,
    FunctionSigContext,
    MethodContext,
    MethodSigContext,
)
from mypy.semanal import SemanticAnalyzer
from mypy.typeanal import TypeAnalyser
from mypy.types import FunctionLike, Instance
from mypy.types import Type as MypyType
from mypy_django_plugin import main
from mypy_django_plugin.django.context import DjangoContext
from mypy_django_plugin.transformers.managers import (
    resolve_manager_method,
    resolve_manager_method_from_instance,
)
from typing_extensions import assert_never

from . import _config, _dependencies, _hook, _known_annotations, _reports, _store, actions


class Hook(
    Generic[_hook.T_Ctx, _hook.T_Ret],
    _hook.Hook["ExtendedMypyStubs", _hook.T_Ctx, _hook.T_Ret],
):
    store: _store.Store

    def extra_init(self) -> None:
        self.store = self.plugin.store


class ExtendedMypyStubs(main.NewSemanalDjangoPlugin):
    """
    The ``ExtendedMypyStubs`` mypy plugin extends the
    ``mypy_django_plugin.main.NewSemanalDjangoPlugin`` found in the active python
    environment.

    It implements the following mypy plugin hooks:

    .. automethod:: get_additional_deps

    .. autoattribute:: get_dynamic_class_hook

    .. autoattribute:: get_type_analyze_hook

    .. autoattribute:: get_function_hook

    .. autoattribute:: get_attribute_hook
    """

    plugin_config: _config.Config

    def __init__(self, options: Options, mypy_version_tuple: tuple[int, int]) -> None:
        super(main.NewSemanalDjangoPlugin, self).__init__(options)
        self.mypy_version_tuple = mypy_version_tuple

        self.plugin_config = _config.Config(options.config_file)
        # Add paths from MYPYPATH env var
        sys.path.extend(mypy_path())
        # Add paths from mypy_path config option
        sys.path.extend(options.mypy_path)

        self.running_in_daemon: bool = "dmypy" in sys.argv[0]

        # Ensure we have a working django context before doing anything
        # So when we try to import things that depend on that, they don't crash us!
        self.django_context = DjangoContext(self.plugin_config.django_settings_module)

        self.report = _reports.Reports.create(
            determine_django_state_script=self.plugin_config.determine_django_state_script,
            django_settings_module=self.plugin_config.django_settings_module,
            scratch_path=self.plugin_config.scratch_path,
        )

        self.store = _store.Store(
            get_model_class_by_fullname=self.django_context.get_model_class_by_fullname,
            lookup_info=self._lookup_info,
            lookup_fully_qualified=self.lookup_fully_qualified,
            django_context_model_modules=self.django_context.model_modules,
            is_installed_model=self._is_installed_model,
            known_concrete_models=self.report.known_concrete_models,
        )

        self.dependencies = _dependencies.Dependencies(
            model_modules=self.store.model_modules,
            report_names_getter=self.report.report_names_getter(
                installed_apps=self.django_context.settings.INSTALLED_APPS,
                model_modules=self.store.model_modules,
                get_model_related_fields=self.django_context.get_model_related_fields,
                get_field_related_model_cls=self.django_context.get_field_related_model_cls,
            ),
        )

    def _is_installed_model(self, instance: Instance) -> bool:
        return self.dependencies.is_model_known(instance.type.fullname)

    def _lookup_info(self, fullname: str) -> TypeInfo | None:
        sym = self.lookup_fully_qualified(fullname)
        if sym and isinstance(sym.node, TypeInfo):
            return sym.node
        else:
            return None

    def _get_symbolnode_for_fullname(
        self, fullname: str, is_function: bool
    ) -> SymbolNode | SymbolTableNode | None:
        sym = self.store.plugin_lookup_fully_qualified(fullname)
        if sym and sym.node:
            return sym.node

        if is_function:
            return None

        if fullname.count(".") < 2:
            return None

        if self._modules is None:
            return None

        # We're on a class and couldn't find the sym, it's likely on a base class
        module, class_name, method_name = fullname.rsplit(".", 2)

        mod = self._modules.get(module)
        if mod is None:
            return None

        class_node = mod.names.get(class_name)
        if not class_node or not isinstance(class_node.node, TypeInfo):
            return None

        for parent in class_node.node.bases:
            if isinstance(parent.type, TypeInfo):
                if isinstance(found := parent.type.names.get(method_name), SymbolTableNode):
                    return found

        return None

    def determine_plugin_version(self, previous_version: int | None = None) -> int:
        """
        Used to set `__version__' where the plugin is defined.

        This lets us tell dmypy to restart itself as necessary.
        """
        if not self.running_in_daemon:
            return 0
        else:
            return self.report.determine_version_hash(
                self.plugin_config.scratch_path, previous_version
            )

    def get_additional_deps(self, file: MypyFile) -> list[tuple[int, str, int]]:
        """
        Ensure that models are re-analyzed if any other models that depend on
        them change.

        We use a generated "report" to re-analyze a file if a new dependency
        is discovered after this file has been processed.
        """
        results = self.dependencies.for_file(
            file.fullname, imports=file.imports, super_deps=super().get_additional_deps(file)
        )
        return results

    @_hook.hook
    class get_dynamic_class_hook(Hook[DynamicClassDefContext, None]):
        """
        This is used to find special methods on the ``Concrete`` class and do appropriate actions.

        For ``Concrete.type_var`` we turn the result into a ``TypeVar`` that can only be one of
        the concrete descendants of the specified class.

        So say we find::

            T_Child = Concrete.type_var("T_Child", Parent)

        Then we turn that into::

            T_Child = TypeVar("T_Child", Child1, Child2, Child3)

        For ``Concrete.cast_as_concrete`` we narrow the target variable to be the concrete equivalent
        of the argument.
        """

        class KnownConcreteMethods(enum.Enum):
            type_var = "type_var"
            cast_as_concrete = "cast_as_concrete"

        method_name: KnownConcreteMethods

        def choose(self) -> bool:
            class_name, _, method_name = self.fullname.rpartition(".")
            try:
                self.method_name = self.KnownConcreteMethods(method_name)
            except ValueError:
                return False
            else:
                info = self.plugin._get_typeinfo_or_none(class_name)
                return bool(info and info.has_base(_known_annotations.KnownClasses.CONCRETE.value))

        def run(self, ctx: DynamicClassDefContext) -> None:
            assert isinstance(ctx.api, SemanticAnalyzer)

            sem_analyzing = actions.SemAnalyzing(self.store, api=ctx.api)

            if self.method_name is self.KnownConcreteMethods.type_var:
                return sem_analyzing.transform_type_var_classmethod(ctx)
            elif self.method_name is self.KnownConcreteMethods.cast_as_concrete:
                return sem_analyzing.transform_cast_as_concrete(ctx)
            else:
                assert_never(self.method_name)

    @_hook.hook
    class get_type_analyze_hook(Hook[AnalyzeTypeContext, MypyType]):
        """
        Resolve classes annotated with ``Concrete`` or ``DefaultQuerySet``.
        """

        annotation: _known_annotations.KnownAnnotations

        def choose(self) -> bool:
            try:
                self.annotation = _known_annotations.KnownAnnotations(self.fullname)
            except ValueError:
                return False
            else:
                return True

        def run(self, ctx: AnalyzeTypeContext) -> MypyType:
            assert isinstance(ctx.api, TypeAnalyser)
            assert isinstance(ctx.api.api, SemanticAnalyzer)

            type_analyzer = actions.TypeAnalyzer(self.store, ctx.api, ctx.api.api)
            return type_analyzer.analyze(ctx, self.annotation)

    @_hook.hook
    class get_attribute_hook(Hook[AttributeContext, MypyType]):
        """
        An implementation of the change found in
        https://github.com/typeddjango/django-stubs/pull/2027
        """

        def choose(self) -> bool:
            return self.super_hook is resolve_manager_method

        def run(self, ctx: AttributeContext) -> MypyType:
            assert isinstance(ctx.api, TypeChecker)

            type_checking = actions.TypeChecking(self.store, api=ctx.api)

            return type_checking.extended_get_attribute_resolve_manager_method(
                ctx, resolve_manager_method_from_instance=resolve_manager_method_from_instance
            )

    class _get_method_or_function_hook(Hook[MethodContext | FunctionContext, MypyType]):
        runner: Callable[[MethodContext | FunctionContext], MypyType | None]

        def extra_init(self) -> None:
            super().extra_init()
            self.shared_logic = actions.SharedModifyReturnTypeLogic(
                self.store,
                fullname=self.fullname,
                get_symbolnode_for_fullname=functools.partial(
                    self.plugin._get_symbolnode_for_fullname,
                    is_function=self.__class__.__name__ == "get_function_hook",
                ),
            )

        def choose(self) -> bool:
            if self.shared_logic.choose():
                self.runner = self.shared_logic.run
                return True
            else:
                return False

        def run(self, ctx: FunctionContext | MethodContext) -> MypyType:
            result = self.runner(ctx)
            if result is not None:
                return result

            if self.super_hook is not None:
                return self.super_hook(ctx)

            return ctx.default_return_type

    @_hook.hook
    class get_method_hook(_get_method_or_function_hook):
        pass

    @_hook.hook
    class get_function_hook(_get_method_or_function_hook):
        pass

    class _get_method_or_function_signature_hook(
        Hook[MethodSigContext | FunctionSigContext, FunctionLike]
    ):
        def extra_init(self) -> None:
            super().extra_init()
            self.shared_logic = actions.SharedCheckTypeGuardsLogic(
                self.store,
                fullname=self.fullname,
                get_symbolnode_for_fullname=functools.partial(
                    self.plugin._get_symbolnode_for_fullname,
                    is_function=self.__class__.__name__ == "get_function_hook",
                ),
            )

        def choose(self) -> bool:
            return self.shared_logic.choose()

        def run(self, ctx: MethodSigContext | FunctionSigContext) -> FunctionLike:
            result = self.shared_logic.run(ctx)
            if result is not None:
                return result

            if self.super_hook is not None:
                return self.super_hook(ctx)

            return ctx.default_signature

    @_hook.hook
    class get_method_signature_hook(_get_method_or_function_signature_hook):
        pass

    @_hook.hook
    class get_function_signature_hook(_get_method_or_function_signature_hook):
        pass



import abc
import dataclasses
from collections.abc import Iterator, MutableMapping
from typing import Final, Protocol

from mypy.checker import TypeChecker
from mypy.nodes import (
    SYMBOL_FUNCBASE_TYPES,
    CallExpr,
    Context,
    Decorator,
    IndexExpr,
    MemberExpr,
    RefExpr,
    SymbolNode,
    SymbolTableNode,
    TypeInfo,
    Var,
)
from mypy.plugin import (
    AttributeContext,
    FunctionContext,
    FunctionSigContext,
    MethodContext,
    MethodSigContext,
)
from mypy.types import (
    AnyType,
    CallableType,
    FunctionLike,
    Instance,
    ProperType,
    TypeOfAny,
    TypeType,
    TypeVarType,
    UnboundType,
    UnionType,
    get_proper_type,
)
from mypy.types import Type as MypyType
from typing_extensions import Self

from .. import _known_annotations, _store
from . import _annotation_resolver


class FailFunc(Protocol):
    def __call__(self, message: str) -> None: ...


class ResolveManagerMethodFromInstance(Protocol):
    def __call__(
        self, instance: Instance, method_name: str, ctx: AttributeContext
    ) -> MypyType: ...


class GetSymbolNode(Protocol):
    def __call__(self, fullname: str) -> SymbolNode | SymbolTableNode | None: ...


TYPING_SELF: Final[str] = "typing.Self"
TYPING_EXTENSION_SELF: Final[str] = "typing_extensions.Self"

_TypeVarMap = MutableMapping[TypeVarType | str, Instance | TypeType | UnionType]


@dataclasses.dataclass
class Finder:
    _api: TypeChecker

    def find_type_vars(
        self, item: MypyType, _chain: list[ProperType] | None = None
    ) -> tuple[list[tuple[bool, TypeVarType]], ProperType]:
        if _chain is None:
            _chain = []

        result: list[tuple[bool, TypeVarType]] = []

        item = get_proper_type(item)

        is_type: bool = False
        if isinstance(item, TypeType):
            is_type = True
            item = item.item

        if isinstance(item, TypeVarType):
            if item.fullname not in [TYPING_EXTENSION_SELF, TYPING_SELF]:
                result.append((is_type, item))

        elif isinstance(item, UnionType):
            for arg in item.items:
                proper = get_proper_type(arg)
                if isinstance(proper, TypeType):
                    proper = proper.item

                if proper not in _chain:
                    _chain.append(proper)
                    for nxt_is_type, nxt in self.find_type_vars(arg, _chain=_chain)[0]:
                        result.append((is_type or nxt_is_type, nxt))

        return result, item

    def determine_if_concrete(
        self, item: ProperType
    ) -> _known_annotations.KnownAnnotations | None:
        concrete_annotation: _known_annotations.KnownAnnotations | None = None

        if isinstance(item, Instance):
            try:
                concrete_annotation = _known_annotations.KnownAnnotations(item.type.fullname)
            except ValueError:
                pass

        return concrete_annotation


@dataclasses.dataclass
class BasicTypeInfo:
    func: CallableType
    fail: FailFunc

    is_type: bool
    is_guard: bool

    item: ProperType
    finder: Finder
    type_vars: list[tuple[bool, TypeVarType]]
    lookup_info: _store.LookupInfo
    concrete_annotation: _known_annotations.KnownAnnotations | None
    unwrapped_type_guard: ProperType | None

    @classmethod
    def create(
        cls,
        func: CallableType,
        fail: FailFunc,
        finder: Finder,
        lookup_info: _store.LookupInfo,
        item: MypyType | None = None,
    ) -> Self:
        is_type: bool = False
        is_guard: bool = False

        item_passed_in: bool = item is not None

        if item is None:
            if func.type_guard:
                is_guard = True
                item = func.type_guard
            else:
                item = func.ret_type

        item = get_proper_type(item)
        if isinstance(item, TypeType):
            is_type = True
            item = item.item

        unwrapped_type_guard: ProperType | None = None
        if isinstance(item, UnboundType) and item.name == "__ConcreteWithTypeVar__":
            unwrapped_type_guard = get_proper_type(item.args[0])
            if is_type:
                unwrapped_type_guard = TypeType(unwrapped_type_guard)

            item = item.args[0]

        item = get_proper_type(item)
        if isinstance(item, TypeType):
            is_type = True
            item = item.item

        concrete_annotation = finder.determine_if_concrete(item)
        if concrete_annotation and not item_passed_in and isinstance(item, Instance):
            type_vars, item = finder.find_type_vars(UnionType(item.args))
        else:
            type_vars, item = finder.find_type_vars(item)

        if isinstance(item, UnionType) and len(item.items) == 1:
            item = item.items[0]

        return cls(
            func=func,
            fail=fail,
            item=get_proper_type(item),
            finder=finder,
            is_type=is_type,
            is_guard=is_guard,
            type_vars=type_vars,
            lookup_info=lookup_info,
            concrete_annotation=concrete_annotation,
            unwrapped_type_guard=unwrapped_type_guard,
        )

    def _clone_with_item(self, item: MypyType) -> Self:
        return self.create(
            func=self.func,
            fail=self.fail,
            item=item,
            finder=self.finder,
            lookup_info=self.lookup_info,
        )

    @property
    def contains_concrete_annotation(self) -> bool:
        if self.concrete_annotation is not None:
            return True

        for item in self.items():
            if item.item is self.item:
                continue
            if item.contains_concrete_annotation:
                return True

        return False

    def items(self) -> Iterator[Self]:
        if isinstance(self.item, UnionType):
            for item in self.item.items:
                yield self._clone_with_item(item)
        else:
            yield self._clone_with_item(self.item)

    def map_type_vars(self, ctx: MethodContext | FunctionContext) -> _TypeVarMap:
        result: _TypeVarMap = {}

        formal_by_name = {arg.name: arg.typ for arg in self.func.formal_arguments()}

        for arg_name, arg_type in zip(ctx.callee_arg_names, ctx.arg_types):
            underlying = get_proper_type(formal_by_name[arg_name])
            if isinstance(underlying, TypeType):
                underlying = underlying.item

            if isinstance(underlying, TypeVarType):
                found_type = get_proper_type(arg_type[0])

                if isinstance(found_type, CallableType):
                    found_type = get_proper_type(found_type.ret_type)

                if isinstance(found_type, TypeType):
                    found_type = found_type.item

                if isinstance(found_type, UnionType):
                    found_type = UnionType(
                        tuple(
                            item
                            if not isinstance(item := get_proper_type(it), TypeType)
                            else item.item
                            for it in found_type.items
                        )
                    )

                if isinstance(found_type, Instance | UnionType):
                    result[underlying] = found_type

        if isinstance(ctx, MethodContext):
            ctx_type = ctx.type
            if isinstance(ctx_type, TypeType):
                ctx_type = ctx_type.item

            if isinstance(ctx.type, CallableType):
                if isinstance(ctx.type.ret_type, Instance | TypeType):
                    ctx_type = ctx.type.ret_type

            if isinstance(ctx_type, TypeType):
                ctx_type = ctx_type.item

            if isinstance(ctx_type, Instance):
                for self_name in [TYPING_EXTENSION_SELF, TYPING_SELF]:
                    result[self_name] = ctx_type

        for is_type, type_var in self.type_vars:
            found: ProperType | None = None
            if type_var in result:
                found = result[type_var]
            else:
                choices = [
                    v
                    for k, v in result.items()
                    if (isinstance(k, TypeVarType) and k.name == type_var.name)
                    or (k == TYPING_SELF and type_var.name == "Self")
                ]
                if len(choices) == 1:
                    result[type_var] = choices[0]
                else:
                    self.fail(f"Failed to find an argument that matched the type var {type_var}")

            if found is not None:
                if is_type:
                    result[type_var] = TypeType(found)

        return result

    def transform(
        self,
        type_checking: "TypeChecking",
        context: Context,
        type_vars_map: _TypeVarMap,
        resolver: _annotation_resolver.AnnotationResolver,
    ) -> Instance | TypeType | UnionType | AnyType | None:
        if self.concrete_annotation is None:
            found: Instance | TypeType | UnionType

            if isinstance(self.item, TypeVarType):
                if self.item in type_vars_map:
                    found = type_vars_map[self.item]
                elif self.item.fullname in [TYPING_EXTENSION_SELF, TYPING_SELF]:
                    found = type_vars_map[self.item.fullname]
                elif self.item.name == "Self" and TYPING_SELF in type_vars_map:
                    found = type_vars_map[TYPING_SELF]
                else:
                    self.fail(f"Failed to work out type for type var {self.item}")
                    return AnyType(TypeOfAny.from_error)
            elif not isinstance(self.item, TypeType | Instance):
                self.fail(f"Got an unexpected item in the concrete annotation, {self.item}")
                return AnyType(TypeOfAny.from_error)
            else:
                found = self.item

            if self.is_type and not isinstance(found, TypeType):
                return TypeType(found)
            else:
                return found

        models: list[Instance | TypeType] = []
        for child in self.items():
            nxt = child.transform(type_checking, context, type_vars_map, resolver=resolver)
            if nxt is None or isinstance(nxt, AnyType | UnionType):
                # Children in self.items() should never return UnionType from transform
                return nxt

            if self.is_type and not isinstance(nxt, TypeType):
                nxt = TypeType(nxt)

            models.append(nxt)

        arg: MypyType
        if len(models) == 1:
            arg = models[0]
        else:
            arg = UnionType(tuple(models))

        return resolver.resolve(self.concrete_annotation, arg)


class TypeChecking:
    def __init__(self, store: _store.Store, *, api: TypeChecker) -> None:
        self.api = api
        self.store = store

    def _named_type_or_none(
        self, fullname: str, args: list[MypyType] | None = None
    ) -> Instance | None:
        node = self.lookup_info(fullname)
        if not isinstance(node, TypeInfo):
            return None
        if args:
            return Instance(node, args)
        return Instance(node, [AnyType(TypeOfAny.special_form)] * len(node.defn.type_vars))

    def _named_type(self, fullname: str, args: list[MypyType] | None = None) -> Instance:
        node = self.lookup_info(fullname)
        assert isinstance(node, TypeInfo)
        if args:
            return Instance(node, args)
        return Instance(node, [AnyType(TypeOfAny.special_form)] * len(node.defn.type_vars))

    def _get_info(self, context: Context) -> BasicTypeInfo | None:
        found: ProperType | None = None
        if isinstance(context, CallExpr):
            found = get_proper_type(self.api.expr_checker.accept(context.callee))
        elif isinstance(context, IndexExpr):
            found = get_proper_type(self.api.expr_checker.accept(context.base))
            if isinstance(found, Instance) and found.args:
                found = get_proper_type(found.args[-1])

        if found is None:
            return None

        if isinstance(found, Instance):
            if not (call := found.type.names.get("__call__")) or not (calltype := call.type):
                return None

            func = get_proper_type(calltype)
        else:
            func = found

        if not isinstance(func, CallableType):
            return None

        return BasicTypeInfo.create(
            func=func,
            fail=lambda msg: self.api.fail(msg, context),
            finder=Finder(_api=self.api),
            lookup_info=self.lookup_info,
        )

    def check_typeguard(self, ctx: MethodSigContext | FunctionSigContext) -> FunctionLike | None:
        info = self._get_info(ctx.context)
        if info is None:
            return None

        if info.is_guard and info.type_vars and info.contains_concrete_annotation:
            # Mypy plugin system doesn't currently provide an opportunity to resolve a type guard when it's for a concrete annotation that uses a type var
            self.api.fail(
                "Can't use a TypeGuard that uses a Concrete Annotation that uses type variables",
                ctx.context,
            )

            if info.unwrapped_type_guard:
                return ctx.default_signature.copy_modified(type_guard=info.unwrapped_type_guard)

        return None

    def modify_return_type(self, ctx: MethodContext | FunctionContext) -> MypyType | None:
        info = self._get_info(ctx.context)
        if info is None:
            return None

        if info.is_guard and info.type_vars and info.concrete_annotation is not None:
            # Mypy plugin system doesn't currently provide an opportunity to resolve a type guard when it's for a concrete annotation that uses a type var
            return None

        if not info.contains_concrete_annotation:
            return None

        type_vars_map = info.map_type_vars(ctx)

        resolver = _annotation_resolver.AnnotationResolver(
            self.store,
            defer=lambda: True,
            fail=lambda msg: ctx.api.fail(msg, ctx.context),
            lookup_info=self.lookup_info,
            named_type_or_none=self._named_type_or_none,
        )

        result = info.transform(self, ctx.context, type_vars_map, resolver=resolver)
        if isinstance(result, UnionType) and len(result.items) == 1:
            return result.items[0]
        else:
            return result

    def extended_get_attribute_resolve_manager_method(
        self,
        ctx: AttributeContext,
        *,
        resolve_manager_method_from_instance: ResolveManagerMethodFromInstance,
    ) -> MypyType:
        """
        Copied from django-stubs after https://github.com/typeddjango/django-stubs/pull/2027

        A 'get_attribute_hook' that is intended to be invoked whenever the TypeChecker encounters
        an attribute on a class that has 'django.db.models.BaseManager' as a base.
        """
        # Skip (method) type that is currently something other than Any of type `implementation_artifact`
        default_attr_type = get_proper_type(ctx.default_attr_type)
        if not isinstance(default_attr_type, AnyType):
            return default_attr_type
        elif default_attr_type.type_of_any != TypeOfAny.implementation_artifact:
            return default_attr_type

        # (Current state is:) We wouldn't end up here when looking up a method from a custom _manager_.
        # That's why we only attempt to lookup the method for either a dynamically added or reverse manager.
        if isinstance(ctx.context, MemberExpr):
            method_name = ctx.context.name
        elif isinstance(ctx.context, CallExpr) and isinstance(ctx.context.callee, MemberExpr):
            method_name = ctx.context.callee.name
        else:
            ctx.api.fail("Unable to resolve return type of queryset/manager method", ctx.context)
            return AnyType(TypeOfAny.from_error)

        if isinstance(ctx.type, Instance):
            return resolve_manager_method_from_instance(
                instance=ctx.type, method_name=method_name, ctx=ctx
            )
        elif isinstance(ctx.type, UnionType) and all(
            isinstance(get_proper_type(instance), Instance) for instance in ctx.type.items
        ):
            items: list[Instance] = []
            for instance in ctx.type.items:
                inst = get_proper_type(instance)
                if isinstance(inst, Instance):
                    items.append(inst)

            resolved = tuple(
                resolve_manager_method_from_instance(
                    instance=inst, method_name=method_name, ctx=ctx
                )
                for inst in items
            )
            return UnionType(resolved)
        else:
            ctx.api.fail(
                f'Unable to resolve return type of queryset/manager method "{method_name}"',
                ctx.context,
            )
            return AnyType(TypeOfAny.from_error)

    def lookup_info(self, fullname: str) -> TypeInfo | None:
        return self.store.plugin_lookup_info(fullname)


class _SharedConcreteAnnotationLogic(abc.ABC):
    def __init__(
        self,
        store: _store.Store,
        fullname: str,
        get_symbolnode_for_fullname: GetSymbolNode,
    ) -> None:
        self.store = store
        self.fullname = fullname
        self.get_symbolnode_for_fullname = get_symbolnode_for_fullname

    def _choose_with_concrete_annotation(self) -> bool:
        sym_node = self.get_symbolnode_for_fullname(self.fullname)
        if not sym_node:
            return False

        if isinstance(sym_node, TypeInfo):
            if "__call__" not in sym_node.names:
                return False
            ret_type = sym_node.names["__call__"].type
        elif isinstance(
            sym_node, (*SYMBOL_FUNCBASE_TYPES, Decorator, SymbolTableNode, Var, RefExpr)
        ):
            ret_type = sym_node.type
        else:
            return False

        ret_type = get_proper_type(ret_type)

        if isinstance(ret_type, CallableType):
            if ret_type.type_guard:
                ret_type = get_proper_type(ret_type.type_guard)
            else:
                ret_type = get_proper_type(ret_type.ret_type)

        if isinstance(ret_type, TypeType):
            ret_type = ret_type.item

        if isinstance(ret_type, UnboundType) and ret_type.name == "__ConcreteWithTypeVar__":
            ret_type = get_proper_type(ret_type.args[0])

        if isinstance(ret_type, Instance):
            try:
                _known_annotations.KnownAnnotations(ret_type.type.fullname)
            except ValueError:
                return False
            else:
                return True
        else:
            return False

    def _type_checking(
        self, ctx: MethodContext | FunctionContext | MethodSigContext | FunctionSigContext
    ) -> TypeChecking:
        assert isinstance(ctx.api, TypeChecker)

        return TypeChecking(self.store, api=ctx.api)


class SharedModifyReturnTypeLogic(_SharedConcreteAnnotationLogic):
    """
    Shared logic for modifying the return type of methods and functions that use a concrete
    annotation with a type variable.

    Note that the signature hook will already raise errors if a concrete annotation is
    used with a type var in a type guard.
    """

    def choose(self) -> bool:
        return self._choose_with_concrete_annotation()

    def run(self, ctx: MethodContext | FunctionContext) -> MypyType | None:
        return self._type_checking(ctx).modify_return_type(ctx)


class SharedCheckTypeGuardsLogic(_SharedConcreteAnnotationLogic):
    """
    Shared logic for modifying the signature of methods and functions.

    This is only used to find cases where a concrete annotation with a type var
    is used in a type guard.

    In this situation the mypy plugin system does not provide an opportunity to fully resolve
    the type guard.
    """

    def choose(self) -> bool:
        return self._choose_with_concrete_annotation()

    def run(self, ctx: MethodSigContext | FunctionSigContext) -> FunctionLike | None:
        return self._type_checking(ctx).check_typeguard(ctx)



import pytest
from pytest_mypy_plugins import MypyPluginsConfig, MypyPluginsScenario

from .scenario import Scenario


@pytest.fixture
def scenario(
    mypy_plugins_config: MypyPluginsConfig, mypy_plugins_scenario: MypyPluginsScenario
) -> Scenario:
    """
    Polish the sharp edges of the pytest mypy plugin
    """
    mypy_plugins_scenario.additional_mypy_config = (
        "\n[mypy.plugins.django-stubs]\n" "django_settings_module = mysettings"
    )
    return Scenario(mypy_plugins_config, mypy_plugins_scenario)



import pytest
from extended_mypy_django_plugin_test_driver import OutputBuilder, Scenario


class TestConcreteAnnotations:
    def test_cast_as_concrete(self, scenario: Scenario) -> None:
        @scenario.run_and_check_mypy_after(installed_apps=["leader", "simple"])
        def _(expected: OutputBuilder) -> None:
            scenario.file(
                expected,
                "main.py",
                """
                from simple.models import Follow1, Follow2
                from leader.models import Leader
                from extended_mypy_django_plugin import Concrete

                model: type[Leader] = Follow1
                # ^ REVEAL ^ type[leader.models.Leader]

                model = Concrete.cast_as_concrete(model)
                # ^ REVEAL ^ Union[type[simple.models.Follow1], type[simple.models.Follow2]]

                model2: type[Leader] = Follow1
                # ^ REVEAL ^ type[leader.models.Leader]

                model3 = Concrete.cast_as_concrete(model2)
                # ^ REVEAL ^ Union[type[simple.models.Follow1], type[simple.models.Follow2]]

                model2
                # ^ REVEAL ^ type[leader.models.Leader]

                instance: Leader = Follow1.objects.create()
                # ^ REVEAL ^ leader.models.Leader

                instance = Concrete.cast_as_concrete(instance)
                # ^ REVEAL ^ Union[simple.models.Follow1, simple.models.Follow2]

                instance2: Leader = Follow1.objects.create()
                # ^ REVEAL ^ leader.models.Leader

                instance3 = Concrete.cast_as_concrete(instance2)
                # ^ REVEAL ^ Union[simple.models.Follow1, simple.models.Follow2]

                instance2
                # ^ REVEAL ^ leader.models.Leader
                """,
            )

    def test_simple_annotation(self, scenario: Scenario) -> None:
        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            scenario.file(
                expected,
                "main.py",
                """
                from extended_mypy_django_plugin import Concrete, DefaultQuerySet
                from typing import cast, TypeGuard

                from myapp.models import Parent, Child1, Child2

                def check_cls_with_type_guard(cls: type[Parent]) -> TypeGuard[type[Concrete[Parent]]]:
                    return True

                def check_instance_with_type_guard(cls: Parent) -> TypeGuard[Concrete[Parent]]:
                    return True

                models: Concrete[Parent]
                # ^ REVEAL ^ Union[myapp.models.Child1, myapp.models.Child2, myapp.models.Child3, myapp2.models.ChildOther]

                qs: DefaultQuerySet[Parent]
                # ^ REVEAL ^ Union[django.db.models.query.QuerySet[myapp.models.Child1, myapp.models.Child1], myapp.models.Child2QuerySet, django.db.models.query.QuerySet[myapp.models.Child3, myapp.models.Child3], django.db.models.query.QuerySet[myapp2.models.ChildOther, myapp2.models.ChildOther]]

                cls: type[Parent] = Child1
                assert check_cls_with_type_guard(cls)
                cls
                # ^ REVEAL ^ Union[type[myapp.models.Child1], type[myapp.models.Child2], type[myapp.models.Child3], type[myapp2.models.ChildOther]]

                instance: Parent = cast(Child1, None)
                assert check_instance_with_type_guard(instance)
                instance
                # ^ REVEAL ^ Union[myapp.models.Child1, myapp.models.Child2, myapp.models.Child3, myapp2.models.ChildOther]

                children: Concrete[Parent]
                # ^ REVEAL ^ Union[myapp.models.Child1, myapp.models.Child2, myapp.models.Child3, myapp2.models.ChildOther]

                children_qs: DefaultQuerySet[Parent]
                # ^ REVEAL ^ Union[django.db.models.query.QuerySet[myapp.models.Child1, myapp.models.Child1], myapp.models.Child2QuerySet, django.db.models.query.QuerySet[myapp.models.Child3, myapp.models.Child3], django.db.models.query.QuerySet[myapp2.models.ChildOther, myapp2.models.ChildOther]]

                child: Concrete[Child1]
                # ^ REVEAL ^ myapp.models.Child1

                child1_qs: DefaultQuerySet[Child1]
                # ^ REVEAL ^ django.db.models.query.QuerySet[myapp.models.Child1, myapp.models.Child1]

                child2_qs: DefaultQuerySet[Child2]
                # ^ REVEAL ^ myapp.models.Child2QuerySet

                t1_children: type[Concrete[Parent]]
                # ^ REVEAL ^ Union[type[myapp.models.Child1], type[myapp.models.Child2], type[myapp.models.Child3], type[myapp2.models.ChildOther]]

                t1_children_qs: type[DefaultQuerySet[Parent]]
                # ^ REVEAL ^ Union[type[django.db.models.query.QuerySet[myapp.models.Child1, myapp.models.Child1]], type[myapp.models.Child2QuerySet], type[django.db.models.query.QuerySet[myapp.models.Child3, myapp.models.Child3]], type[django.db.models.query.QuerySet[myapp2.models.ChildOther, myapp2.models.ChildOther]]]

                t1_child: type[Concrete[Child1]]
                # ^ REVEAL ^ type[myapp.models.Child1]

                t1_child1_qs: type[DefaultQuerySet[Child1]]
                # ^ REVEAL ^ type[django.db.models.query.QuerySet[myapp.models.Child1, myapp.models.Child1]]

                t1_child2_qs: type[DefaultQuerySet[Child2]]
                # ^ REVEAL ^ type[myapp.models.Child2QuerySet]

                t2_children: Concrete[type[Parent]]
                # ^ REVEAL ^ Union[type[myapp.models.Child1], type[myapp.models.Child2], type[myapp.models.Child3], type[myapp2.models.ChildOther]]

                t2_children_qs: DefaultQuerySet[type[Parent]]
                # ^ REVEAL ^ Union[type[django.db.models.query.QuerySet[myapp.models.Child1, myapp.models.Child1]], type[myapp.models.Child2QuerySet], type[django.db.models.query.QuerySet[myapp.models.Child3, myapp.models.Child3]], type[django.db.models.query.QuerySet[myapp2.models.ChildOther, myapp2.models.ChildOther]]]

                t2_child: Concrete[type[Child1]]
                # ^ REVEAL ^ type[myapp.models.Child1]

                t2_child1_qs: DefaultQuerySet[type[Child1]]
                # ^ REVEAL ^ type[django.db.models.query.QuerySet[myapp.models.Child1, myapp.models.Child1]]

                t2_child2_qs: DefaultQuerySet[type[Child2]]
                # ^ REVEAL ^ type[myapp.models.Child2QuerySet]
                """,
            )

    def test_can_use_type_var_before_class_is_defined(self, scenario: Scenario) -> None:
        @scenario.run_and_check_mypy_after(installed_apps=["example"])
        def _(expected: OutputBuilder) -> None:
            scenario.file(expected, "example/__init__.py", "")

            scenario.file(
                expected,
                "example/apps.py",
                """
                from django.apps import AppConfig

                class Config(AppConfig):
                    name = "example"
                """,
            )

            scenario.file(
                expected,
                "example/models.py",
                """
                from __future__ import annotations

                from django.db import models
                from typing_extensions import Self
                from extended_mypy_django_plugin import Concrete

                T_Leader = Concrete.type_var("T_Leader", "Leader")

                class Leader(models.Model):
                    @classmethod
                    def new(cls) -> Concrete[Self]:
                        cls = Concrete.cast_as_concrete(cls)
                        # ^ REVEAL ^ Union[type[example.models.Follower1], type[example.models.Follower2]]
                        return cls.objects.create()

                    class Meta:
                        abstract = True


                class Follower1(Leader):
                    ...

                class Follower2(Leader):
                    ...

                def make_instance(cls: type[T_Leader]) -> T_Leader:
                    return cls.new()
                """,
            )

            scenario.file(
                expected,
                "main.py",
                """
                from example.models import Leader, Follower1, Follower2, make_instance
                from extended_mypy_django_plugin import Concrete

                instance1 = make_instance(Follower1)
                # ^ REVEAL ^ example.models.Follower1

                instance2 = make_instance(Follower2)
                # ^ REVEAL ^ example.models.Follower2
                """,
                # TODO: Make this possible
                # would require generating overload variants on make_instance
                # model: type[Concrete[Leader]] = Follower1
                # # ^ REVEAL ^ Union[type[example.models.Follower1], type[example.models.Follower2]]
                # instance3 = make_instance(model)
                # # ^ REVEAL ^ Union[example.models.Follower1, example.models.Follower2]
                # """,
            )

    def test_using_concrete_annotation_on_class_used_in_annotation(
        self, scenario: Scenario
    ) -> None:
        @scenario.run_and_check_mypy_after(installed_apps=["example"])
        def _(expected: OutputBuilder) -> None:
            scenario.file(expected, "example/__init__.py", "")

            scenario.file(
                expected,
                "example/apps.py",
                """
                from django.apps import AppConfig

                class Config(AppConfig):
                    name = "example"
                """,
            )

            scenario.file(
                expected,
                "example/models.py",
                """
                from __future__ import annotations

                from django.db import models
                from collections.abc import Callable
                from typing_extensions import Self
                from typing import Protocol
                from extended_mypy_django_plugin import Concrete, DefaultQuerySet

                class Leader(models.Model):
                    @classmethod
                    def new(cls) -> Concrete[Self]:
                        cls = Concrete.cast_as_concrete(cls)
                        # ^ REVEAL ^ Union[type[example.models.Follower1], type[example.models.Follower2]]
                        return cls.objects.create()

                    def qs(self) -> DefaultQuerySet[Self]:
                        self = Concrete.cast_as_concrete(self)
                        # ^ REVEAL ^ Union[example.models.Follower1, example.models.Follower2]
                        return self.__class__.objects.filter(pk=self.pk)

                    class Meta:
                        abstract = True

                class Follower1QuerySet(models.QuerySet["Follower1"]):
                    ...

                Follower1Manager = models.Manager.from_queryset(Follower1QuerySet)

                class Follower1(Leader):
                    objects = Follower1Manager()

                class Follower2(Leader):
                    ...
                """,
            )

            scenario.file(
                expected,
                "main.py",
                """
                from example.models import Leader, Follower1, Follower2
                from extended_mypy_django_plugin import Concrete

                model: type[Leader] = Follower1
                # ^ REVEAL ^ type[example.models.Leader]

                leader = model.new()
                # ^ REVEAL ^ Union[example.models.Follower1, example.models.Follower2]

                qs = leader.qs()
                # ^ REVEAL ^ Union[example.models.Follower1QuerySet, django.db.models.query.QuerySet[example.models.Follower2, example.models.Follower2]]

                follower1 = Follower1.new()
                # ^ REVEAL ^ example.models.Follower1

                qs3 = Follower2.new().qs()
                # ^ REVEAL ^ django.db.models.query.QuerySet[example.models.Follower2, example.models.Follower2]

                other: Leader = Follower1.new()
                # ^ REVEAL ^ example.models.Leader

                narrowed = Concrete.cast_as_concrete(other)
                # ^ REVEAL ^ Union[example.models.Follower1, example.models.Follower2]
                other
                # ^ REVEAL ^ example.models.Leader
                """,
            )

    def test_resolving_concrete_type_vars(self, scenario: Scenario) -> None:
        @scenario.run_and_check_mypy_after(installed_apps=["example"])
        def _(expected: OutputBuilder) -> None:
            scenario.file(expected, "example/__init__.py", "")

            scenario.file(
                expected,
                "example/apps.py",
                """
                from django.apps import AppConfig

                class Config(AppConfig):
                    name = "example"
                """,
            )

            scenario.file(
                expected,
                "example/models.py",
                """
                from __future__ import annotations

                from django.db import models
                from collections.abc import Callable
                from typing_extensions import Self
                from typing import Protocol
                from extended_mypy_django_plugin import Concrete, DefaultQuerySet

                class Leader(models.Model):
                    class Meta:
                        abstract = True

                # TODO: the error is fixed in future commit
                T_Leader = Concrete.type_var("T_Leader", Leader)
                # ^ ERROR(misc) ^ No concrete children found for example.models.Leader

                class Follower1QuerySet(models.QuerySet["Follower1"]):
                    ...

                Follower1Manager = models.Manager.from_queryset(Follower1QuerySet)

                class Follower1(Leader):
                    objects = Follower1Manager()

                class Follower2(Leader):
                    ...

                def make_queryset(instance: T_Leader, /) -> DefaultQuerySet[T_Leader]:
                    return instance.__class__.objects.filter(pk=instance.pk)

                class MakeQueryset(Protocol):
                    def __call__(self, instance: T_Leader, /) -> DefaultQuerySet[T_Leader]:
                        ...

                functions: list[MakeQueryset] = [make_queryset]

                functions2: list[Callable[[T_Leader], DefaultQuerySet[T_Leader]]] = [make_queryset] # type: ignore[list-item]
                """,
            )

            scenario.file(
                expected,
                "main.py",
                """
                from example.models import Leader, Follower1, Follower2, functions, functions2, make_queryset
                from extended_mypy_django_plugin import Concrete

                follower1 = Follower1.objects.create()
                # ^ REVEAL ^ example.models.Follower1

                func = functions[0]
                # ^ REVEAL ^ example.models.MakeQueryset

                func2 = functions2[0]
                
                qs1 = func(follower1)
                # ^ REVEAL ^ example.models.Follower1QuerySet

                qs2 = func2(follower1)
                # ^ REVEAL ^ example.models.Follower1QuerySet

                qs5 = make_queryset(follower1)
                # ^ REVEAL ^ example.models.Follower1QuerySet
                """,
            )

    def test_restarts_dmypy_if_names_of_known_settings_change(self, scenario: Scenario) -> None:
        if not scenario.for_daemon:
            pytest.skip("Test only relevant for the daemon")

        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            pass

        # New names should restart
        custom_settings = """
        ONE: int = 1
        TWO: str = "2"
        """

        @scenario.run_and_check_mypy_after(
            additional_properties={"custom_settings": custom_settings}
        )
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_restart()

        @scenario.run_and_check_mypy_after(
            additional_properties={"custom_settings": custom_settings}
        )
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_not_restart()

        # Same names, different types should  restart
        custom_settings = """
        ONE: str = "1"
        TWO: str = "2"
        """

        @scenario.run_and_check_mypy_after(
            additional_properties={"custom_settings": custom_settings}
        )
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_restart()

        @scenario.run_and_check_mypy_after(
            additional_properties={"custom_settings": custom_settings}
        )
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_not_restart()

        # Same names, same types, different values should not restart
        custom_settings = """
        ONE: str = "3"
        TWO: str = "4"
        """

        @scenario.run_and_check_mypy_after(
            additional_properties={"custom_settings": custom_settings}
        )
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_not_restart()

        @scenario.run_and_check_mypy_after(
            additional_properties={"custom_settings": custom_settings}
        )
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_not_restart()

        # removed names, same types, different values should not restart
        custom_settings = """
        TWO: str = "4"
        """

        @scenario.run_and_check_mypy_after(
            additional_properties={"custom_settings": custom_settings}
        )
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_restart()

        @scenario.run_and_check_mypy_after(
            additional_properties={"custom_settings": custom_settings}
        )
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_not_restart()

    def test_sees_apps_removed_when_they_still_exist_but_no_longer_installed(
        self, scenario: Scenario
    ) -> None:
        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            scenario.file(
                expected,
                "main.py",
                """
                from extended_mypy_django_plugin import Concrete, DefaultQuerySet

                from myapp.models import Parent

                models: Concrete[Parent]
                # ^ REVEAL[concrete-parent] ^ Union[myapp.models.Child1, myapp.models.Child2, myapp.models.Child3, myapp2.models.ChildOther]

                qs: DefaultQuerySet[Parent]
                # ^ REVEAL[qs-parent] ^ Union[django.db.models.query.QuerySet[myapp.models.Child1, myapp.models.Child1], myapp.models.Child2QuerySet, django.db.models.query.QuerySet[myapp.models.Child3, myapp.models.Child3], django.db.models.query.QuerySet[myapp2.models.ChildOther, myapp2.models.ChildOther]]
                """,
            )

        # Now let's remove myapp2 from the installed_apps and see that the daemon restarts and myapp2 is removed from the revealed types

        @scenario.run_and_check_mypy_after(installed_apps=["myapp"])
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_restart()

            (
                expected.on("main.py")
                .remove_from_revealed_type(
                    "concrete-parent",
                    ", myapp2.models.ChildOther",
                )
                .remove_from_revealed_type(
                    "qs-parent",
                    ", django.db.models.query.QuerySet[myapp2.models.ChildOther, myapp2.models.ChildOther]",
                )
            )

    def test_does_not_see_apps_that_exist_but_are_not_installed(self, scenario: Scenario) -> None:
        @scenario.run_and_check_mypy_after(
            installed_apps=["myapp"], copied_apps=["myapp", "myapp2"]
        )
        def _(expected: OutputBuilder) -> None:
            scenario.file(
                expected,
                "main.py",
                """
                from extended_mypy_django_plugin import Concrete, DefaultQuerySet

                from myapp.models import Parent

                model: Concrete[Parent]
                model.concrete_from_myapp

                qs: DefaultQuerySet[Parent]
                qs.values("concrete_from_myapp")
                """,
            )

        # And after installing the app means the types expand

        @scenario.run_and_check_mypy_after(installed_apps=["myapp", "myapp2"])
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_restart()

            (
                expected.on("main.py")
                .add_error(
                    6,
                    "union-attr",
                    'Item "ChildOther" of "Child1 | Child2 | Child3 | ChildOther" has no attribute "concrete_from_myapp"',
                )
                .add_error(
                    9,
                    "misc",
                    "Cannot resolve keyword 'concrete_from_myapp' into field. Choices are: concrete_from_myapp2, id, one, two",
                )
            )

    def test_sees_models_when_they_are_added_and_installed(self, scenario: Scenario) -> None:
        @scenario.run_and_check_mypy_after(installed_apps=["myapp"])
        def _(expected: OutputBuilder) -> None:
            scenario.file(
                expected,
                "main.py",
                """
                from extended_mypy_django_plugin import Concrete, DefaultQuerySet

                from myapp.models import Parent

                models: Concrete[Parent]
                # ^ REVEAL[concrete-parent] ^ Union[myapp.models.Child1, myapp.models.Child2, myapp.models.Child3]

                qs: DefaultQuerySet[Parent]
                qs.values("concrete_from_myapp")
                # ^ TAG[concrete-qs] ^
                """,
            )

        # and the models become available after being installed

        @scenario.run_and_check_mypy_after(installed_apps=["myapp", "myapp2"])
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_restart()

            (
                expected.on("main.py")
                .add_revealed_type(
                    "concrete-parent",
                    "Union[myapp.models.Child1, myapp.models.Child2, myapp.models.Child3, myapp2.models.ChildOther]",
                )
                .add_error(
                    "concrete-qs",
                    "misc",
                    "Cannot resolve keyword 'concrete_from_myapp' into field. Choices are: concrete_from_myapp2, id, one, two",
                )
            )

        # And same output if nothing changes

        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_not_restart()

    def test_sees_new_models(self, scenario: Scenario) -> None:
        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            scenario.file(
                expected,
                "main.py",
                """
                from extended_mypy_django_plugin import Concrete, DefaultQuerySet

                from myapp.models import Parent

                models: Concrete[Parent]
                models.two

                qs: DefaultQuerySet[Parent]
                qs.values("two")
                """,
            )

        # And if we add some more models

        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            scenario.append_to_file(
                expected,
                "myapp2/models.py",
                """
                class Another(Parent):
                    pass
                """,
            )

            expected.daemon_should_restart()

            (
                expected.on("main.py")
                .add_error(
                    6,
                    "union-attr",
                    'Item "Another" of "Child1 | Child2 | Child3 | Another | ChildOther" has no attribute "two"',
                )
                .add_error(
                    9, "misc", "Cannot resolve keyword 'two' into field. Choices are: id, one"
                )
            )

        # And the new model remains after a rerun

        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_not_restart()

    def test_sees_changes_in_custom_querysets_within_app(self, scenario: Scenario) -> None:
        @scenario.run_and_check_mypy_after(installed_apps=["leader", "follower1"])
        def _(expected: OutputBuilder) -> None:
            scenario.file(
                expected,
                "main.py",
                """
                from extended_mypy_django_plugin import Concrete, DefaultQuerySet

                from leader.models import Leader

                models: Concrete[Leader]
                # ^ REVEAL[concrete-leader] ^ follower1.models.follower1.Follower1

                qs: DefaultQuerySet[Leader]
                # ^ REVEAL[qs-leader] ^ follower1.models.follower1.Follower1QuerySet

                qs.good_ones().values("nup")
                # ^ ERROR(misc)[error1] ^ Cannot resolve keyword 'nup' into field. Choices are: from_follower1, good, id
                """,
            )

        # And then add another model

        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            scenario.file(
                expected,
                "follower1/models/__init__.py",
                """
                from .follower1 import Follower1
                from .follower2 import Follower2

                __all__ = ["Follower1", "Follower2"]
                """,
            )

            scenario.file(
                expected,
                "follower1/models/follower2.py",
                """
                from django.db import models
                from leader.models import Leader

                class Follower2QuerySet(models.QuerySet["Follower2"]):
                    def good_ones(self) -> "Follower2QuerySet":
                        return self.filter(good=True)

                Follower2Manager = models.Manager.from_queryset(Follower2QuerySet)

                class Follower2(Leader):
                    good = models.BooleanField()

                    objects = Follower2Manager()
                """,
            )

            expected.daemon_should_restart()

            (
                expected.on("main.py")
                .add_revealed_type(
                    "concrete-leader",
                    "Union[follower1.models.follower1.Follower1, follower1.models.follower2.Follower2]",
                )
                .add_revealed_type(
                    "qs-leader",
                    "Union[follower1.models.follower1.Follower1QuerySet, follower1.models.follower2.Follower2QuerySet]",
                )
                .replace_errors(
                    "error1",
                    (
                        "misc",
                        "Cannot resolve keyword 'nup' into field. Choices are: from_follower1, good, id",
                    ),
                    (
                        "misc",
                        "Cannot resolve keyword 'nup' into field. Choices are: good, id",
                    ),
                )
            )

        # And same output, no daemon restart on no change

        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_not_restart()

        # And removing the method from the queryset

        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            scenario.file(
                expected,
                "follower1/models/follower2.py",
                """
                from django.db import models
                from leader.models import Leader

                class Follower2QuerySet(models.QuerySet["Follower2"]):
                    pass

                Follower2Manager = models.Manager.from_queryset(Follower2QuerySet)

                class Follower2(Leader):
                    good = models.BooleanField()

                    objects = Follower2Manager()
                """,
            )

            (
                expected.on("main.py").replace_errors(
                    "error1",
                    (
                        "union-attr",
                        'Item "Follower2QuerySet" of "Follower1QuerySet | Follower2QuerySet" has no attribute "good_ones"',
                    ),
                    (
                        "misc",
                        "Cannot resolve keyword 'nup' into field. Choices are: from_follower1, good, id",
                    ),
                )
            )

    def test_sees_changes_in_custom_querysets_in_new_apps(self, scenario: Scenario) -> None:
        @scenario.run_and_check_mypy_after(installed_apps=["leader", "follower1"])
        def _(expected: OutputBuilder) -> None:
            scenario.file(
                expected,
                "main.py",
                """
                from extended_mypy_django_plugin import Concrete, DefaultQuerySet

                from leader.models import Leader

                models: Concrete[Leader]
                # ^ REVEAL[concrete-leader] ^ follower1.models.follower1.Follower1

                qs: DefaultQuerySet[Leader]
                # ^ REVEAL[qs-leader] ^ follower1.models.follower1.Follower1QuerySet

                qs.good_ones().values("nup")
                # ^ ERROR(misc)[error1] ^ Cannot resolve keyword 'nup' into field. Choices are: from_follower1, good, id
                """,
            )

        # Let's then add a new app with new models

        @scenario.run_and_check_mypy_after(installed_apps=["leader", "follower1", "follower2"])
        def _(expected: OutputBuilder) -> None:
            scenario.file(expected, "follower2/__init__.py", "")

            scenario.file(
                expected,
                "follower2/apps.py",
                """
                from django.apps import AppConfig
                class Config(AppConfig):
                    name = "follower2"
                """,
            )

            scenario.file(
                expected,
                "follower2/models.py",
                """
                from django.db import models
                from leader.models import Leader

                class Follower2QuerySet(models.QuerySet["Follower2"]):
                    def good_ones(self) -> "Follower2QuerySet":
                        return self.filter(good=True)

                Follower2Manager = models.Manager.from_queryset(Follower2QuerySet)

                class Follower2(Leader):
                    good = models.BooleanField()

                    objects = Follower2Manager()
                """,
            )

            expected.daemon_should_restart()

            (
                expected.on("main.py")
                .add_revealed_type(
                    "concrete-leader",
                    "Union[follower1.models.follower1.Follower1, follower2.models.Follower2]",
                )
                .add_revealed_type(
                    "qs-leader",
                    "Union[follower1.models.follower1.Follower1QuerySet, follower2.models.Follower2QuerySet]",
                )
                .replace_errors(
                    "error1",
                    (
                        "misc",
                        "Cannot resolve keyword 'nup' into field. Choices are: from_follower1, good, id",
                    ),
                    (
                        "misc",
                        "Cannot resolve keyword 'nup' into field. Choices are: good, id",
                    ),
                )
            )

        # And everything stays the same on rerun

        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            expected.daemon_should_not_restart()

        # And it sees where custom queryset gets queryset manager removed

        @scenario.run_and_check_mypy_after
        def _(expected: OutputBuilder) -> None:
            scenario.file(
                expected,
                "follower2/models.py",
                """
                from django.db import models
                from leader.models import Leader

                class Follower2QuerySet(models.QuerySet["Follower2"]):
                    pass

                Follower2Manager = models.Manager.from_queryset(Follower2QuerySet)

                class Follower2(Leader):
                    good = models.BooleanField()

                    objects = Follower2Manager()
                """,
            )

            (
                expected.on("main.py").replace_errors(
                    "error1",
                    (
                        "union-attr",
                        'Item "Follower2QuerySet" of "Follower1QuerySet | Follower2QuerySet" has no attribute "good_ones"',
                    ),
                    (
                        "misc",
                        "Cannot resolve keyword 'nup' into field. Choices are: from_follower1, good, id",
                    ),
                )
            )



from mypy.nodes import (
    GDEF,
    AssignmentStmt,
    CastExpr,
    NameExpr,
    StrExpr,
    SymbolTableNode,
    TypeInfo,
    TypeVarExpr,
    Var,
)
from mypy.plugin import AnalyzeTypeContext, DynamicClassDefContext
from mypy.semanal import SemanticAnalyzer
from mypy.typeanal import TypeAnalyser
from mypy.types import (
    AnyType,
    CallableType,
    Instance,
    TypeOfAny,
    TypeType,
    TypeVarType,
    UnboundType,
    UnionType,
    get_proper_type,
)
from mypy.types import (
    Type as MypyType,
)

from .. import _known_annotations, _store
from . import _annotation_resolver


class TypeAnalyzer:
    def __init__(self, store: _store.Store, api: TypeAnalyser, sem_api: SemanticAnalyzer) -> None:
        self.api = api
        self.store = store
        self.sem_api = sem_api

    def _lookup_info(self, fullname: str) -> TypeInfo | None:
        instance = self.sem_api.named_type_or_none(fullname)
        if instance:
            return instance.type

        return self.store.plugin_lookup_info(fullname)

    def analyze(
        self, ctx: AnalyzeTypeContext, annotation: _known_annotations.KnownAnnotations
    ) -> MypyType:
        def defer() -> bool:
            if self.sem_api.final_iteration:
                return True
            else:
                self.sem_api.defer()
                return False

        resolver = _annotation_resolver.AnnotationResolver(
            self.store,
            defer=defer,
            fail=lambda msg: self.api.fail(msg, ctx.context),
            lookup_info=self._lookup_info,
            named_type_or_none=self.sem_api.named_type_or_none,
        )

        type_arg, rewrap = resolver.find_type_arg(ctx.type, self.api.analyze_type)
        if type_arg is None:
            return ctx.type

        if rewrap:
            info = self._lookup_info(annotation.value)
            if info is None:
                self.api.fail(f"Couldn't find information for {annotation.value}", ctx.context)
                return ctx.type

            if isinstance(type_arg, TypeType) and isinstance(type_arg.item, TypeVarType):
                if type_arg.item.fullname == "extended_mypy_django_plugin.annotations.T_Parent":
                    return ctx.type
            elif isinstance(type_arg, TypeVarType):
                if type_arg.fullname == "extended_mypy_django_plugin.annotations.T_Parent":
                    return ctx.type

            return UnboundType(
                "__ConcreteWithTypeVar__",
                [Instance(info, [type_arg])],
                line=ctx.context.line,
                column=ctx.context.column,
            )

        resolved = resolver.resolve(annotation, type_arg)
        if resolved is None:
            return ctx.type
        else:
            return resolved


class SemAnalyzing:
    def __init__(self, store: _store.Store, *, api: SemanticAnalyzer) -> None:
        self.api = api
        self.store = store

    def _lookup_info(self, fullname: str) -> TypeInfo | None:
        instance = self.api.named_type_or_none(fullname)
        if instance:
            return instance.type

        return self.store.plugin_lookup_info(fullname)

    def transform_cast_as_concrete(self, ctx: DynamicClassDefContext) -> None:
        if len(ctx.call.args) != 1:
            self.api.fail("Concrete.cast_as_concrete takes exactly one argument", ctx.call)
            return None

        node = self.api.lookup_type_node(ctx.call.args[0])
        if not node or not node.node:
            self.api.fail("Failed to lookup the argument", ctx.call)
            return None

        arg_node = self.api.lookup_current_scope(node.node.name)

        if arg_node is None or arg_node.type is None or arg_node.node is None:
            return None

        if not isinstance(arg_node.node, Var):
            return None

        arg_node_typ = get_proper_type(arg_node.type)

        is_type: bool = False
        if isinstance(arg_node_typ, TypeType):
            is_type = True
            arg_node_typ = arg_node_typ.item

        if not isinstance(arg_node_typ, Instance | UnionType | TypeVarType):
            self.api.fail(
                f"Unsure what to do with the type of the argument given to cast_as_concrete: {arg_node_typ}",
                ctx.call,
            )
            return None

        if isinstance(arg_node_typ, TypeVarType):
            if (
                self.api.is_func_scope()
                and isinstance(self.api.type, TypeInfo)
                and arg_node_typ.name == "Self"
            ):
                replacement: Instance | TypeType | None = self.api.named_type_or_none(
                    self.api.type.fullname
                )
                if replacement:
                    arg_node_typ = replacement
                    if is_type:
                        replacement = TypeType(replacement)

                    if self.api.scope.function and isinstance(
                        self.api.scope.function.type, CallableType
                    ):
                        if self.api.scope.function.type.arg_names[0] == ctx.name:
                            # Avoid getting an assignment error trying to assign a union of the concrete types to
                            # a variable typed in terms of Self
                            self.api.scope.function.type.arg_types[0] = replacement
                else:
                    self.api.fail("Failed to resolve Self", ctx.call)
                    return None
            else:
                self.api.fail(
                    f"Resolving type variables for cast_as_concrete not implemented: {arg_node_typ}",
                    ctx.call,
                )
                return None

        def defer() -> bool:
            if self.api.final_iteration:
                return True
            else:
                self.api.defer()
                return False

        resolver = _annotation_resolver.AnnotationResolver(
            self.store,
            defer=defer,
            fail=lambda msg: self.api.fail(msg, ctx.call),
            lookup_info=self._lookup_info,
            named_type_or_none=self.api.named_type_or_none,
        )

        concrete = resolver.resolve(
            _known_annotations.KnownAnnotations.CONCRETE,
            TypeType(arg_node_typ) if is_type else arg_node_typ,
        )
        if not concrete:
            return None

        ctx.call.analyzed = CastExpr(ctx.call.args[0], concrete)
        ctx.call.analyzed.line = ctx.call.line
        ctx.call.analyzed.column = ctx.call.column
        ctx.call.analyzed.accept(self.api)

        return None

    def transform_type_var_classmethod(self, ctx: DynamicClassDefContext) -> None:
        if len(ctx.call.args) != 2:
            self.api.fail("Concrete.type_var takes exactly two arguments", ctx.call)
            return None

        name = self.api.extract_typevarlike_name(
            AssignmentStmt([NameExpr(ctx.name)], ctx.call.callee), ctx.call
        )
        if name is None:
            return None

        second_arg = ctx.call.args[1]
        model_name: str

        if isinstance(second_arg, StrExpr):
            if second_arg.value.isidentifier():
                model_name = f"{self.api.cur_mod_id}.{second_arg.value}"
            else:
                self.api.fail(
                    "Concrete.type_var needs to take the instance or name of a model class in this file",
                    ctx.call,
                )
                return None
        elif isinstance(second_arg, NameExpr):
            model_name = f"{self.api.cur_mod_id}.{second_arg.name}"
        else:
            self.api.fail(
                "Second argument to type_var must be either a simple name or a string of the name",
                ctx.call,
            )

        parent = self.api.lookup_fully_qualified_or_none(model_name)

        if parent is None:
            if self.api.final_iteration:
                self.api.fail(
                    f"Failed to locate the model provided to to make {ctx.name} ({model_name})",
                    ctx.call,
                )
                return None
            else:
                self.api.defer()
                return None

        if not isinstance(parent.node, TypeInfo):
            self.api.fail(
                "Second argument to Concrete.type_var was not pointing at a class", ctx.call
            )
            return

        object_type = self.api.named_type("builtins.object")
        values = self.store.retrieve_concrete_children_types(
            parent.node, self._lookup_info, self.api.named_type_or_none
        )
        if not values:
            self.api.fail(f"No concrete children found for {parent.node.fullname}", ctx.call)

        type_var_expr = TypeVarExpr(
            name=name,
            fullname=f"{self.api.cur_mod_id}.{name}",
            values=list(values),
            upper_bound=object_type,
            default=AnyType(TypeOfAny.from_omitted_generics),
        )

        module = self.api.modules[self.api.cur_mod_id]
        module.names[name] = SymbolTableNode(GDEF, type_var_expr, plugin_generated=True)
        return None
